# -*- coding: utf-8 -*-

import requests
import fourch
from .thread import thread


class board(object):
    """ fourch.board is the master instance which allows easy access to the creation
        of thread objects.
    """

    def __init__(self, name, https=False, urls=None):
        """ Create the board instance, and initialize variables.

            :param name: The board name, minus slashes. e.g., 'b', 'x', 'tv'
            :param https: Should we use HTTPS or HTTP?
            :type https: bool
            :param urls: the url dictionary, with all the required url bases
            :type urls: dict or None
        """
        self.name = name
        self.https = https
        self._urls = urls or fourch.urls
        self._session = requests.Session()
        self._session.headers.update({"User-Agent": "fourch/{0} (@https://github.com/plausibility/4ch)".format(fourch.__version__)})
        self._cache = {}  # id: fourch.thread, this stores prefetched threads

    @property
    def _base_url(self):
        """Create the base API url"""
        return self._proto + self._urls["api"]

    @property
    def _proto(self):
        """ Decide which HTTP protocol to use (HTTPS or HTTP) and return it, along with slashes.

            :return: the protocol to use
            :rtype: str
        """
        return "https://" if self.https else "http://"

    def catalog(self):
        """ Get a list of all the thread OPs and last replies.
        """
        url = self._base_url + self._urls["api_catalog"].format(board=self.name)
        r = self._session.get(url)
        return r.json()

    def threads(self):
        """ Get a list of all the threads alive, and which page they're on.

            You can cross-reference this with a threads number to see which page
            it's on at the time of calling.
        """
        url = self._base_url + self._urls["api_threads"].format(board=self.name)
        r = self._session.get(url)
        return r.json()

    def thread(self, res, update_cache=True):
        """ Create a :class:`fourch.thread` object.
            If the thread has already been fetched, return the cached thread.

            :param res: the thread number to fetch
            :type res: str or int
            :param update_cache: should we update if it's cached?
            :type update_cache: bool
            :return: the :class:`fourch.thread` object
            :rtype: :class:`fourch.thread` or None
        """
        if res in self._cache:
            t = self._cache[res]
            if update_cache:
                t.update()
            return t

        url = self._base_url + self._urls["api_thread"].format(board=self.name, thread=res)

        r = self._session.get(url)
        t = thread.from_req(self, res, r)
        if t is not None:
            self._cache[res] = t
        return t

    def page(self, page=1, update_each=False):
        """ Return all the threads in a single page.
            The page number is now one-indexed. The first page is 1, second is 2, etc.

            If a thread has already been cached, return the cache entry rather than making a new thread.

            :param page: page to pull threads from
            :type page: int
            :param update_each: should each thread be updated, to pull all replies
            :type update_each: bool
            :return: a list of :class:`fourch.thread` objects, corresponding to all threads on given page
            :rtype: list
        """
        url = self._base_url + self._urls["api_board"].format(board=self.name, page=page)
        r = self._session.get(url)
        if r.status_code != requests.codes.ok:
            r.raise_for_status()

        json = r.json()
        threads = []

        for thj in json["threads"]:
            t = None
            res = thj["posts"][0]["no"]

            if res in self._cache:
                t = self._cache[res]
                t._should_update = True
            else:
                t = thread.from_json(self, thj, last_modified=r.headers["last-modified"])
                self._cache[res] = t

            if update_each:
                t.update()
            threads.append(t)

        return threads

    def thread_exists(self, res):
        """ Figure out whether or not a thread exists.
            This is as easy as checking if it 404s.

            :param res: the thread number to fetch
            :type res: str or int
            :return: whether or not the given thread exists
            :rtype: bool
        """
        url = self._base_url + self._urls["api_thread"].format(board=self.name, thread=res)
        return self._session.head(url).status_code == requests.codes.ok



# -*- coding: utf-8 -*-
from ._version import __version__

urls = {
    "api": "a.4cdn.org",
    "boards": "boards.4chan.org",
    "images": "i.4cdn.org",
    "thumbs": "t.4cdn.org",

    # These are tacked to the end of the api url after formatting.
    "api_board": "/{board}/{page}.json",
    "api_thread": "/{board}/thread/{thread}.json",
    "api_threads": "/{board}/threads.json",
    "api_catalog": "/{board}/catalog.json",
    "api_boards": "/boards.json"
}


class struct:
    def __init__(self, **entries):
        self.__dict__.update(entries)



# -*- coding: utf-8 -*-

import requests
from .reply import reply


class thread(object):
    """ This object stores information about the given thread.
        It has a list of fourch.replies, as well as options to
        easily pull in updates (new posts), and create an instance
        with the json of a thread.
    """

    def __init__(self, board, res):
        """ Create the thread instance and initialize variables.

            :param board: the :class:`fourch.board` parent instance
            :type board: :class:`fourch.board`
            :param res: the given threads number
            :type res: str or int
        """
        self._board = board
        self.res = res
        self.alive = True  # has the thread died on us? 404 = True
        self.op = None
        self.replies = []
        self.omitted_posts = 0
        self.omitted_images = 0
        self._should_update = False  # If this is a precached thread, should it get updated?
        self._last_modified = None  # HTTP Last-Modified header for If-Modified-Since

    def __repr__(self):
        end = ""
        if self.omitted_posts or self.omitted_images:
            end = ", {0} omitted posts, {1} omitted images".format(self.omitted_posts, self.omitted_images)
        return "<{0}.{1} /{2}/{3}, {4} replies{5}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            self._board.name,
            self.res,
            len(self.replies),
            end
        )

    @staticmethod
    def from_req(board, res, r):
        """ Create a thread object from the given request.
            If the thread has 404d, this will return None,
            and if it isn't 200 OK, it will raise_for_status().
            Actually creates the thread by calling :func:`from_json`.

            :param board: the :class:`fourch.board` parent instance
            :type board: :class:`fourch.board`
            :param res: the given threads number
            :type res: str or int
            :param r: the requests object
            :type r: requests.models.Response
        """
        if r.status_code == requests.codes.not_found:
            return None
        elif r.status_code == requests.codes.ok:
            return thread.from_json(board, r.json(), res=res, last_modified=r.headers["last-modified"])
        else:
            r.raise_for_status()

    @staticmethod
    def from_json(board, json, res=None, last_modified=None):
        """ Create a thread object from the given JSON data.

            :param board: the :class:`fourch.board` parent instance
            :type board: :class:`fourch.board`
            :param json: the json data from the 4chan API
            :type board: dict
            :param res: the given threads number
            :type res: str or int
            :param last_modified: when was the page last modified
            :type last_modified: int or None
            :return: the created :class:`fourch.thread`
            :rtype: :class:`fourch.thread`
        """
        t = thread(board, res)
        t._last_modified = last_modified

        replies = json["posts"]

        t.op = reply(t, replies.pop(0))
        t.replies = [reply(t, r) for r in replies]

        if res is None:
            t._should_update = True
            t.res = t.op.number
            t.omitted_posts = t.op._json.get("omitted_posts", 0)
            t.omitted_images = t.op._json.get("omitted_images", 0)

        return t

    @property
    def sticky(self):
        """ Is this thread stuck?

            :return: whether or not the thread is stuck
            :rtype: bool
        """
        return self.op.sticky

    @property
    def closed(self):
        """ Is the thread closed?

            :return: whether or not the thread is closed
            :rtype: bool
        """
        return self.op.closed

    @property
    def last_reply(self):
        """ Return the last :class:`fourch.reply` to the thread, or the op
            if there are no replies.

            :return: the last :class:`fourch.reply` to the thread.
            :rtype: :class:`fourch.reply`
        """
        if not self.replies:
            return self.op
        return self.replies[-1]

    @property
    def images(self):
        """ Create a generator which yields all of the image urls for the thread.

            :return: a generator yieling all image urls
            :rtype: generator
        """
        yield self.op.file.url
        for r in self.replies:
            if not r.has_file:
                continue
            yield r.file.url

    def update(self, force=False):
        """ Update the thread, pulling in new replies,
            appending them to the reply pool.

            :param force: should replies be replaced with fresh reply objects
            :type force: bool
            :return: the number of new replies
            :rtype: int
        """
        if not self.alive and not force:
            return 0

        url = self._board._base_url + self._board._urls["api_thread"].format(board=self._board.name, thread=self.res)
        headers = None
        if self._last_modified:
            # If-Modified-Since, to not waste bandwidth.
            headers = {
                "If-Modified-Since": self._last_modified
            }

        r = self._board._session.get(url, headers=headers)

        if r.status_code == requests.codes.not_modified:
            # 304 Not Modified
            return 0

        elif r.status_code == requests.codes.not_found:
            # 404 Not Found
            self.alive = False
            # Remove from cache.
            self._board._cache.pop(self.res, None)
            return 0

        elif r.status_code == requests.codes.ok:
            if not self.alive:
                self.alive = True
                self._board._cache[self.res] = self

            self._should_update = False
            self.omitted_posts = 0
            self.omitted_images = 0

            self._last_modified = r.headers["last-modified"]
            replies = r.json()["posts"]

            post_count = len(self.replies)
            self.op = reply(self, replies.pop(0))
            if not force:
                self.replies.extend([reply(self, p) for p in replies if p["no"] > self.last_reply.number])
            else:
                self.replies = [reply(self, p) for p in replies]
            post_count_new = len(self.replies)
            post_count_diff = post_count_new - post_count
            if not post_count_diff:
                return 0
            return post_count_diff

        else:
            r.raise_for_status()



# -*- coding: utf-8 -*-

import fourch
import base64
import re


class reply(object):
    """ This object stores information regarding a specific post
        on any given thread. It uses python properties to easily
        allow access to information.
    """

    def __init__(self, thread, json):
        """ Initialize the reply with the relevant information

            :param thread: the :class:`fourch.thread` parent instance
            :type thread: :class:`fourch.thread`
            :param json: the json data for this post
            :type json: dict
        """
        self._thread = thread
        self._json = json

    def __repr__(self):
        return "<{0}.{1} /{2}/{3}#{4}, image: {5}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            self._thread._board.name,
            self._thread.res,
            self.number,
            bool(self.has_file)
        )

    @property
    def is_op(self):
        """Is this post the OP (first post in thread)"""
        return self._json.get("resto", 1) == 0

    @property
    def number(self):
        """The number relating to this post"""
        return self._json.get("no", 0)

    @property
    def reply_to(self):
        """What post ID is this a reply to"""
        return self._json.get("resto", 0)

    @property
    def sticky(self):
        """Is this thread stuck?"""
        return bool(self._json.get("sticky", 0))

    @property
    def closed(self):
        """Is this thread closed?"""
        return bool(self._json.get("closed", 0))

    @property
    def now(self):
        """Humanized date string of post time"""
        return self._json.get("now", "")

    @property
    def timestamp(self):
        """The UNIX timestamp of post time"""
        return self._json.get("time", 0)

    @property
    def tripcode(self):
        """Trip code, if any, of the post"""
        return self._json.get("trip", "")

    @property
    def id(self):
        """Post ID, if any. (Admin, Mod, Developer, etc)"""
        return self._json.get("id", "")

    @property
    def capcode(self):
        """Post capcode, if any. (none, mod, admin, etc)"""
        return self._json.get("capcode", "")

    @property
    def country(self):
        """The country code this was posted from. Two characters, XX if unknown"""
        return self._json.get("country", "XX")

    @property
    def country_name(self):
        """The name of the country this was posted from"""
        return self._json.get("country_name", "")

    @property
    def email(self):
        """The email attached to the post"""
        return self._json.get("email", "")

    @property
    def subject(self):
        """The subject of the post"""
        return self._json.get("sub", "")

    @property
    def comment(self):
        """The comment, including escaped HTML"""
        return self._json.get("com", "")

    @property
    def comment_text(self):
        """ The stripped (mostly) plain text version of the comment.
            The comment goes through various regexes to become (mostly) clean.

            Some HTML will still be present, this is because Python's :mod:`HTMLParser` won't
            escape everything, and since it's undocumented, only god may know
            how to add more escapes.
        """
        import HTMLParser
        com = self.comment
        # <span class="quote">&gt;text!</span>
        # --- >text!
        com = re.sub("\<span[^>]+\>(?:&gt;|>)([^</]+)\<\/span\>", ">\\1", com, flags=re.I)
        # <a class="quotelink" href="XX#pYYYY">&gt;&gt;YYYY</a>
        # --- >>YYYY
        com = re.sub("\<a[^>]+\>(?:&gt;|>){2}(\d+)\<\/a\>", ">>\\1", com, flags=re.I)
        # Add (OP) to quotelinks to op
        com = re.sub("\>\>({0})".format(self._thread.op.number), ">>\\1 (OP)", com, flags=re.I)
        # <br> or <br /> to newline
        com = re.sub("\<br ?\/?\>", "\n", com, flags=re.I)
        # Send the remaining HTML through the HTMLParser to unescape.
        com = HTMLParser.HTMLParser().unescape(com)
        return com

    @property
    def url(self):
        """The URL of the post on the parent thread"""
        return "{0}{1}/{2}/thread/{3}#p{4}".format(
            self._thread._board._proto,
            self._thread._board._urls["boards"],
            self._thread._board.name,
            self._thread.res,
            self.number
        )

    ## File related
    @property
    def has_file(self):
        """Whether or not this post has an image attached"""
        return "filename" in self._json

    @property
    def file(self):
        """ This holds the information regarding the image attached
            to a post, if there is one at all.

            It returns the relevant information in a class format,
            accessible via ``r.file.url``, for example.

            Information stored:

            - renamed
            - name
            - extension
            - size
            - md5
            - md5b64
            - width
            - height
            - thumb_width
            - thumb_height
            - deleted
            - spoiler
            - url
            - thumb_url

            :return: a struct with information related to image
        """
        if not self.has_file:
            return fourch.struct()
        f = {
            "renamed": self._json.get("tim", 0),
            "name": self._json.get("filename", ""),
            "extension": self._json.get("ext", ""),
            "size": self._json.get("fsize", 0),
            "md5": base64.b64decode(self._json.get("md5")),
            "md5b64": self._json.get("md5", ""),
            "width": self._json.get("w", 0),
            "height": self._json.get("h", 0),
            "thumb_width": self._json.get("tn_w", 0),
            "thumb_height": self._json.get("tn_h", 0),
            "deleted": bool(self._json.get("filedeleted", 0)),
            "spoiler": bool(self._json.get("spoiler", 0)),
            "url": "",
            "thumb_url": ""
        }
        f["url"] = "{0}{1}/{2}/{3}{4}".format(
            self._thread._board._proto,
            fourch.urls["images"],
            self._thread._board.name,
            f["renamed"],
            f["extension"]
        )
        f["thumb_url"] = "{0}{1}/{2}/{3}s.jpg".format(
            self._thread._board._proto,
            fourch.urls["thumbs"],
            self._thread._board.name,
            f["renamed"]
        )
        return fourch.struct(**f)



""" fourch (stylised as 4ch) is an easy-to-implement Python wrapper for
    4chan's JSON API, as provided by moot.

    It uses the documentation of the 4chan API located at: https://github.com/4chan/4chan-API
    
    This is based off of the API last updated March 22, 2013.
    (4chan-API commit: 9a76db1162173cc56f04bf34e09b38d4376314c3)
"""
from ._version import __version__

from .fourch import *

from .thread import thread as thread
from .board import board as board
from .reply import reply as reply
from .boards import boards as boards



# -*- coding: utf-8 -*-

import requests
import fourch


class listed_board:
    """ This object stores info about a board on 4chan.
        It is used by fourch.boards(), a function to
        get a list of all the boards and their properties.
    """

    def __init__(self, json):
        """ Initialize the board with the information it has.

            :param json: the json data for this post
            :type json: dict
        """
        self._json = json

    @property
    def board(self):
        """The board name, minus slashes. e.g., 'b', 'x', 'tv'"""
        return self._json['board']

    @property
    def title(self):
        """The title of the board, e.g., 'Random', 'Paranormal', 'Television & Film'"""
        return self._json['title']

    @property
    def ws(self):
        """Is the board worksafe?"""
        return bool(self._json['ws_board'])

    @property
    def pages(self):
        """How many pages does the board have?"""
        return self._json['pages']

    @property
    def per_page(self):
        """How many posts are there on 1 page?"""
        return self._json['per_page']


def boards(https=False):
    """ Get a list of all boards on 4chan, in :class:`fourch.boards.listed_board` objects
        containing information about the boards.

        :param https: Should we use HTTPS or HTTP?
        :type https: bool
    """
    s = requests.Session()
    s.headers.update({"User-Agent": "fourch/{0} (@https://github.com/plausibility/4ch)".format(fourch.__version__)})
    proto = "https://" if https else "http://"
    url = proto + fourch.urls['api'] + fourch.urls["api_boards"]
    r = s.get(url)
    if r.status_code != requests.codes.ok:
        r.raise_for_status()

    boards = []
    for json_board in r.json()['boards']:
        boards.append(listed_board(json_board))
    return boards



"""fourch
======

.. _docs: https://4ch.readthedocs.org
.. _repo: https://github.com/plausibility/4ch

fourch (stylized as 4ch) is a wrapper to the 4chan JSON API, provided by moot. It allows you to interact with 4chan (in a READONLY way) easily through your scripts.

Originally <strike>stolen</strike> forked from `e000/py-4chan <https://github.com/e000/py-4chan>`_, but then I moved repos and renamed stuff since I'm pretty bad about that.

Requirements
------------

- Python 2.7 (what I test with, 2.x might work)
- requests

Notes
-----

- This isn't guaranteed to work all the time; after all, the API may change, and 4ch will have to be updated accordingly.
- If a feature is missing, open an issue on the `repo`_, and it may well be implemented.

Running / Usage
---------------

- Install & import: ``$ pip install 4ch``, ``import fourch``
- See the `docs`_

Contributing
------------
If you're interested in contributing to the usability of 4ch, or just want to give away stars, you can visit the 4ch github `repo`_.
"""
from setuptools import setup

if __name__ != "__main__":
    import sys
    sys.exit(1)

kw = {
    "name": "4ch",
    "version": "0.4.0",
    "description": "Python wrapper for the 4chan JSON API.",
    "long_description": __doc__,
    "url": "https://github.com/plausibility/4ch",
    "author": "plausibility",
    "author_email": "chris@gibsonsec.org",
    "license": "MIT",
    "packages": ['fourch'],
    "install_requires": ["requests"],
    "zip_safe": False,
    "keywords": "wrapper 4chan chan json",
    "classifiers": [
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2"
    ]
}

if __name__ == "__main__":
    setup(**kw)
